/*
 * libsysactivity
 * http://sourceforge.net/projects/libsysactivity/
 * Copyright (c) 2009 Carlos Olmedo Escobar <carlos.olmedo.e@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <errno.h>
#include <string.h>
#include <inttypes.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>

#include "libsysactivity.h"

static int iterate_devices(struct sa_data_storage* dst, uint16_t dst_size, uint16_t* count);
static int is_a_real_device(char* dev_name);
static int read_stat_file(struct sa_data_storage* dst, char* dev_name);

__thread DIR* dir_sys_block;

int sa_open_data_storage() {
	dir_sys_block = opendir("/sys/block/");
	if (dir_sys_block == NULL)
		return EIO;
	return 0;
}

int sa_close_data_storage() {
	if (dir_sys_block != NULL)
		closedir(dir_sys_block);
	return 0;
}

int sa_count_data_storages(uint16_t* number) {
	if (number == NULL)
		return EINVAL;

	if (iterate_devices(NULL, 0, number) != 0)
		return EIO;
	return 0;
}

int sa_get_data_storage(char* name, struct sa_data_storage* dst) {
	if (name == NULL || dst == NULL)
		return EINVAL;

	if (!is_a_real_device(name))
		return ENODEV;
	return read_stat_file(dst, name);
}

int sa_get_data_storages(struct sa_data_storage* dst, uint16_t dst_size, uint16_t* written) {
	if (dst == NULL || dst_size == 0 || written == NULL)
		return EINVAL;

	return iterate_devices(dst, dst_size, written);
}

static int iterate_devices(struct sa_data_storage* dst, uint16_t dst_size, uint16_t* count) {
	*count = 0;
	rewinddir(dir_sys_block);
	struct dirent* device_folder;
	while ((device_folder = readdir(dir_sys_block)) != NULL) {
		if (!is_a_real_device(device_folder->d_name))
			continue;

		if (dst != NULL) {
			if (*count == dst_size)
				return ENODEV;
			if (read_stat_file(&dst[*count], device_folder->d_name) != 0)
				return EIO;
		}
		(*count)++;
	}

	return 0;
}

static int is_a_real_device(char* dev_name) {
	char line_buffer[64];
	DIR* tmp_dir;

	if (snprintf(line_buffer, sizeof line_buffer, "/sys/block/%s/device/", dev_name) >= sizeof line_buffer)
		return ENODEV;
	tmp_dir = opendir(line_buffer);
	if (tmp_dir == NULL)
		return 0;

	closedir(tmp_dir);
	return 1;
}

static int read_stat_file(struct sa_data_storage* dst, char* dev_name) {
	char line_buffer[192];
	FILE* file_stat;

	if (snprintf(line_buffer, sizeof line_buffer, "/sys/block/%s/stat", dev_name) >= sizeof line_buffer)
		return ENODEV;
	file_stat = fopen(line_buffer, "r");
	if (file_stat == NULL)
		return EIO;
	if (fgets(line_buffer, sizeof line_buffer, file_stat) == NULL)
		return EIO;

	if (sscanf(line_buffer,
			"%"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" ", &dst->reads,
			&dst->reads_merged, &dst->sectors_read, &dst->time_spent_reading, &dst->writes,
			&dst->sectors_written, &dst->time_spent_writing) != 7)
		return EIO;
	strncpy(dst->name, dev_name, sizeof dst->name);
	fclose(file_stat);
	return 0;
}
