/*
 * libsysactivity
 * http://sourceforge.net/projects/libsysactivity/
 * Copyright (c) 2009 Carlos Olmedo Escobar <carlos.olmedo.e@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <stdlib.h>
#include <errno.h>

#include "libsysactivity.h"

static void get_value(char* line, uint64_t* value, int8_t key_length);
static int calculate_swap(struct sa_swap* dst, char* line_buffer);

__thread FILE* file_meminfo;
__thread FILE* file_swaps;

int sa_open_memory() {
	file_swaps = NULL;

	file_meminfo = fopen("/proc/meminfo", "r");
	if (file_meminfo == NULL)
		return EIO;
	file_swaps = fopen("/proc/swaps", "r");
	if (file_swaps == NULL)
		return EIO;
	return 0;
}

int sa_close_memory() {
	if (file_meminfo != NULL)
		fclose(file_meminfo);
	if (file_swaps != NULL)
		fclose(file_swaps);
	return 0;
}

int sa_get_memory(struct sa_memory* dst) {
	if (dst == NULL)
		return EINVAL;

	char line_buffer[64];

	if (fflush(file_meminfo) != 0)
		return EIO;
	rewind(file_meminfo);

	while (fgets(line_buffer, sizeof(line_buffer), file_meminfo) != NULL) {
		if (strncmp(line_buffer, "MemTotal:", sizeof("MemTotal:") - 1) == 0)
			get_value(line_buffer, &dst->total, sizeof("MemTotal:") - 1);
		else if (strncmp(line_buffer, "MemFree:", sizeof("MemFree:") - 1) == 0)
			get_value(line_buffer, &dst->free, sizeof("MemFree:") - 1);
		else if (strncmp(line_buffer, "Active:", sizeof("Active:") - 1) == 0)
			get_value(line_buffer, &dst->active, sizeof("Active:") - 1);
		else if (strncmp(line_buffer, "Inactive:", sizeof("Inactive:") - 1) == 0)
			get_value(line_buffer, &dst->inactive, sizeof("Inactive:") - 1);
		else if (strncmp(line_buffer, "Buffers:", sizeof("Buffers:") - 1) == 0)
			get_value(line_buffer, &dst->buffers, sizeof("Buffers:") - 1);
		else if (strncmp(line_buffer, "SwapTotal:", sizeof("SwapTotal:") - 1) == 0)
			get_value(line_buffer, &dst->swap_total, sizeof("SwapTotal:") - 1);
		else if (strncmp(line_buffer, "SwapFree:", sizeof("SwapFree:") - 1) == 0)
			get_value(line_buffer, &dst->swap_free, sizeof("SwapFree:") - 1);
		else if (strncmp(line_buffer, "SwapCached:", sizeof("SwapCached:") - 1) == 0)
			get_value(line_buffer, &dst->swap_cached, sizeof("SwapCached:") - 1);
		else if (strncmp(line_buffer, "Cached:", sizeof("Cached:") - 1) == 0)
			get_value(line_buffer, &dst->cached, sizeof("Cached:") - 1);
		else if (strncmp(line_buffer, "Dirty:", sizeof("Dirty:") - 1) == 0)
			get_value(line_buffer, &dst->dirty, sizeof("Dirty:") - 1);
	}
	return 0;
}

int sa_count_swaps(uint16_t* number) {
	if (number == NULL)
		return EINVAL;

	if (fflush(file_swaps) != 0)
		return EIO;
	if (fseek(file_swaps, 37, SEEK_SET))
		return EIO;

	char line_buffer[1024];
	*number = 0;
	while (fgets(line_buffer, sizeof(line_buffer), file_swaps) != NULL) {
		(*number)++;
	}
	return 0;
}

int sa_get_swap(uint16_t index, struct sa_swap* dst) {
	if (dst == NULL)
		return EINVAL;

	if (fflush(file_swaps) != 0)
		return EIO;
	if (fseek(file_swaps, 37, SEEK_SET))
		return EIO;

	int i;
	char line_buffer[1024];
	for (i = 0; fgets(line_buffer, sizeof(line_buffer), file_swaps) != NULL; i++) {
		if (i != index)
			continue;

		if (calculate_swap(dst, line_buffer) == ENOSYS)
			return ENOSYS;

		return 0;
	}
	return ENODEV;
}

int sa_get_swaps(struct sa_swap* dst, uint16_t dst_size, uint16_t* written) {
	if (dst == NULL || dst_size == 0 || written == NULL)
		return EINVAL;

	if (fflush(file_swaps) != 0)
		return EIO;
	if (fseek(file_swaps, 37, SEEK_SET))
		return EIO;

	*written = 0;
	int i;
	char line_buffer[1024];
	for (i = 0; i < dst_size; i++) {
		if (fgets(line_buffer, sizeof line_buffer, file_swaps) == NULL)
			return EIO;

		if (calculate_swap(dst, line_buffer) == ENOSYS)
			return ENOSYS;

		(*written)++;
	}
	return 0;
}

static void get_value(char* line, uint64_t* value, int8_t key_length) {
	sscanf(&line[key_length], "%"PRIu64" kB", value);
	*value *= 1024;
}

static int calculate_swap(struct sa_swap* dst, char* line_buffer) {
	char type[16];
	if (sscanf(line_buffer, "%511s %15s %"PRIu64" %"PRIu64" ", dst->name, type, &dst->total,
			&dst->free) < 3)
		return ENOSYS;
	dst->total *= 1024;
	dst->free *= 1024;
	dst->type = strncmp(type, "file", 4) == 0 ? 2 : 1;
	return 0;
}
