/*
 * libsysactivity
 * http://sourceforge.net/projects/libsysactivity/
 * Copyright (c) 2009 Carlos Olmedo Escobar <carlos.olmedo.e@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <inttypes.h>
#include <sys/sysinfo.h>

#include "libsysactivity.h"

static int read_stats(struct sa_process* dst, pid_t pid);

__thread DIR* dir_proc;

int sa_open_process() {
	dir_proc = opendir("/proc/");
	if (dir_proc == NULL)
		return EIO;

	return 0;
}

int sa_close_process() {
	if (dir_proc != NULL)
		closedir(dir_proc);
	return 0;
}

int sa_count_processes(pid_t* number) {
	if (number == NULL)
		return EINVAL;

	struct sysinfo info;
	if (sysinfo(&info) == -1)
		return errno;

	*number = (pid_t) info.procs;
	return 0;
}

int sa_get_processes_ids(pid_t* dst, pid_t dst_size, pid_t* written) {
	if (dst == NULL || dst_size <= 0 || written == NULL)
		return EINVAL;

	rewinddir(dir_proc);
	*written = 0;

	struct dirent* entry;
	pid_t pid;
	while ((entry = readdir(dir_proc)) != NULL) {
		pid = atoi(entry->d_name);
		if (pid == 0)
			continue;
		if (*written == dst_size)
			return ENOMEM;

		dst[*written] = pid;
		(*written)++;
	}

	return 0;
}

int sa_get_process(pid_t pid, struct sa_process* dst) {
	if (pid == 0 || dst == NULL)
		return EINVAL;

	return read_stats(dst, pid);
}

int sa_get_processes(struct sa_process* dst, pid_t dst_size, pid_t* written) {
	if (dst == NULL || dst_size <= 0 || written == NULL)
		return EINVAL;

	rewinddir(dir_proc);
	*written = 0;

	struct dirent* entry;
	int ret;
	pid_t pid;
	while ((entry = readdir(dir_proc)) != NULL) {
		pid = atoi(entry->d_name);
		if (pid == 0)
			continue;
		if (*written == dst_size)
			return ENOMEM;

		ret = read_stats(&dst[*written], pid);
		if (ret != 0) {
			if (ret == ESRCH)
				continue;
			return ret;
		}
		(*written)++;
	}

	return 0;
}

static int read_stats(struct sa_process* dst, pid_t pid) {
	char line_buffer[256];
	FILE* file;

	sprintf(line_buffer, "/proc/%d/stat", pid);
	file = fopen(line_buffer, "r");
	if (file == NULL)
		return ESRCH;
	if (fgets(line_buffer, sizeof line_buffer, file) == NULL)
		return EIO;

	char state;
	if (sscanf(
			line_buffer,
			"%d %255s %c %d %d %d %d %*d %*d %*d %*d %*d %*d %"PRIu64" %"PRIu64" %*d %*d %*d %c %d %*d %"PRIu64" %u %u ",
			&dst->pid, dst->filename, &state, &dst->parent_pid, &dst->pgrp, &dst->sid, &dst->tty,
			&dst->user_time, &dst->sys_time, &dst->nice, &dst->threads, &dst->start_time,
			&dst->vm_size, &dst->rss) != 14)
		return EIO;
	fclose(file);
	switch (state) {
	case 'S':
		dst->state = SLEEPING;
		break;
	case 'D':
		dst->state = SLEEPING_UNINTERRUPTIBLE;
		break;
	case 'Z':
		dst->state = ZOMBIE;
		break;
	case 'T':
		dst->state = STOPPED;
		break;
	default:
		dst->state = RUNNING;
	}

	sprintf(line_buffer, "/proc/%d/cmdline", pid);
	file = fopen(line_buffer, "r");
	if (file == NULL)
		return ESRCH;
	fgets(dst->cmdline, sizeof dst->cmdline, file); // sometimes this file is empty
	fclose(file);

	sprintf(line_buffer, "/proc/%d/status", pid);
	file = fopen(line_buffer, "r");
	if (file == NULL)
		return ESRCH;
	dst->uid = dst->gid = 0;
	while (fgets(line_buffer, sizeof line_buffer, file) != NULL) {
		if (sscanf(line_buffer, "Uid: %d %*d %*d %*d", &dst->uid) == 1)
			continue;
		if (sscanf(line_buffer, "Gid: %d %*d %*d %*d", &dst->gid) == 1)
			break;
	}
	fclose(file);

	return 0;
}
